from typing import Any
from graia.broadcast.entities.event import BaseEvent, BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from pydantic.main import BaseModel


class ApplicationLaunched(BaseEvent):
    app: Any

    def __init__(self, app) -> None:
        super().__init__(app=app)

    class Dispatcher(BaseDispatcher):
        @staticmethod
        def catch(interface: "DispatcherInterface"):
            from .. import GraiaMiraiApplication

            if interface.annotation is GraiaMiraiApplication:
                return interface.event.app


class ApplicationLaunchedBlocking(BaseEvent):
    app: Any

    def __init__(self, app) -> None:
        super().__init__(app=app)

    class Dispatcher(BaseDispatcher):
        @staticmethod
        def catch(interface: "DispatcherInterface"):
            from .. import GraiaMiraiApplication

            if interface.annotation is GraiaMiraiApplication:
                return interface.event.app


class ApplicationShutdowned(BaseEvent):
    app: Any

    def __init__(self, app) -> None:
        super().__init__(app=app)

    class Dispatcher(BaseDispatcher):
        @staticmethod
        def catch(interface: "DispatcherInterface"):
            from .. import GraiaMiraiApplication

            if interface.annotation is GraiaMiraiApplication:
                return interface.event.app



from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from graia.application.message.chain import MessageChain


class MessageChainCatcher(BaseDispatcher):
    @staticmethod
    def catch(interface: "DispatcherInterface"):
        if interface.annotation is MessageChain:
            return interface.event.messageChain



from typing import Callable, Optional, Union
from graia.broadcast.exceptions import ExecutionStop
from graia.broadcast.interrupt.waiter import Waiter
from graia.application.group import Group, Member
from graia.application.message.elements.internal import Quote, Source
from graia.application.message import BotMessage
from graia.application.event.messages import FriendMessage, GroupMessage, TempMessage
from graia.application.friend import Friend


def GroupMessageInterrupt(
    special_group: Optional[Union[Group, int]] = None,
    special_member: Optional[Union[Member, int]] = None,
    quote_access: Optional[Union[BotMessage, Source]] = None,
    custom_judgement: Optional[Callable[[GroupMessage], bool]] = None,
    block_propagation: bool = False,
):
    @Waiter.create_using_function([GroupMessage], block_propagation=block_propagation)
    def GroupMessageInterruptWaiter(event: GroupMessage):
        if special_group:
            if event.sender.group.id != (
                special_group.id if isinstance(special_group, Group) else special_group
            ):
                raise ExecutionStop()
        if special_member:
            if event.sender.id != (
                special_member.id
                if isinstance(special_member, Member)
                else special_member
            ):
                raise ExecutionStop()
        if quote_access:
            quotes = event.messageChain.get(Quote)
            if not quotes:
                raise ExecutionStop()
            quote: Quote = quotes[0]
            if isinstance(quote_access, BotMessage):
                if quote.id != quote_access.messageId:
                    raise ExecutionStop()
            elif isinstance(quote_access, Source):
                if quote.id != quote_access.id:
                    raise ExecutionStop()
        if custom_judgement:
            if not custom_judgement(event):
                raise ExecutionStop()
        return event

    return GroupMessageInterruptWaiter


def FriendMessageInterrupt(
    special_friend: Optional[Union[Friend, int]] = None,
    quote_access: Optional[Union[BotMessage, Source]] = None,
    custom_judgement: Optional[Callable[[FriendMessage], bool]] = None,
    block_propagation: bool = False,
):
    @Waiter.create_using_function([FriendMessage], block_propagation=block_propagation)
    def FriendMessageInterruptWaiter(event: FriendMessage):
        if special_friend:
            if event.sender.id != (
                special_friend.id
                if isinstance(special_friend, Friend)
                else special_friend
            ):
                raise ExecutionStop()
        if quote_access:
            quotes = event.messageChain.get(Quote)
            if not quotes:
                raise ExecutionStop()
            quote: Quote = quotes[0]
            if isinstance(quote_access, BotMessage):
                if quote.id != quote_access.messageId:
                    raise ExecutionStop()
            elif isinstance(quote_access, Source):
                if quote.id != quote_access.id:
                    raise ExecutionStop()
        if custom_judgement:
            if not custom_judgement(event):
                raise ExecutionStop()
        return event

    return FriendMessageInterruptWaiter


def TempMessageInterrupt(
    special_group: Optional[Union[Group, int]] = None,
    special_member: Optional[Union[Member, int]] = None,
    quote_access: Optional[Union[BotMessage, Source]] = None,
    custom_judgement: Optional[Callable[[TempMessage], bool]] = None,
    block_propagation: bool = False,
):
    @Waiter.create_using_function([TempMessage], block_propagation=block_propagation)
    def TempMessageInterruptWaiter(event: TempMessage):
        if special_group:
            if event.sender.group.id != (
                special_group.id if isinstance(special_group, Group) else special_group
            ):
                raise ExecutionStop()
        if special_member:
            if event.sender.id != (
                special_member.id
                if isinstance(special_member, Member)
                else special_member
            ):
                raise ExecutionStop()
        if quote_access:
            quotes = event.messageChain.get(Quote)
            if not quotes:
                raise ExecutionStop()
            quote: Quote = quotes[0]
            if isinstance(quote_access, BotMessage):
                if quote.id != quote_access.messageId:
                    raise ExecutionStop()
            elif isinstance(quote_access, Source):
                if quote.id != quote_access.id:
                    raise ExecutionStop()
        if custom_judgement:
            if not custom_judgement(event):
                raise ExecutionStop()
        return event

    return TempMessageInterruptWaiter



from pydantic import validator

from graia.application.event.dispatcher import MessageChainCatcher
from graia.application.message.chain import MessageChain
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from graia.application.group import Member, Group
from graia.application.friend import Friend
from graia.application.message.elements.internal import Source
from . import ApplicationDispatcher, MiraiEvent


class SourceElementDispatcher(BaseDispatcher):
    @staticmethod
    def catch(interface: DispatcherInterface):
        if interface.annotation is Source:
            return interface.event.messageChain.getFirst(Source)


class FriendMessage(MiraiEvent):
    type: str = "FriendMessage"
    messageChain: MessageChain
    sender: Friend

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainCatcher, ApplicationDispatcher, SourceElementDispatcher]

        @staticmethod
        def catch(interface: DispatcherInterface):
            if interface.annotation is Friend:
                return interface.event.sender


class GroupMessage(MiraiEvent):
    type: str = "GroupMessage"
    messageChain: MessageChain
    sender: Member

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainCatcher, ApplicationDispatcher, SourceElementDispatcher]

        @staticmethod
        def catch(interface: DispatcherInterface):
            if interface.annotation is Group:
                return interface.event.sender.group
            elif interface.annotation is Member:
                return interface.event.sender


class TempMessage(MiraiEvent):
    type: str = "TempMessage"
    messageChain: MessageChain
    sender: Member

    @classmethod
    def parse_obj(cls, obj):
        return super().parse_obj(obj)

    class Dispatcher(BaseDispatcher):
        mixin = [MessageChainCatcher, ApplicationDispatcher, SourceElementDispatcher]

        @staticmethod
        def catch(interface: DispatcherInterface):
            if interface.annotation is Group:
                return interface.event.sender.group
            elif interface.annotation is Member:
                return interface.event.sender



import functools
from typing import Any, Callable, ContextManager, Iterable, List, Union, TypeVar
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from graia.application.event import MiraiEvent
from .exceptions import (
    AccountMuted,
    AccountNotFound,
    InvaildArgument,
    InvaildAuthkey,
    InvaildSession,
    NotSupportedVersion,
    TooLongMessage,
    UnauthorizedSession,
    UnknownTarget,
)
from .context import enter_context
import inspect

_T = TypeVar("_T")


def applicationContextManager(func: Callable):
    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        with enter_context(app=self):
            return await func(self, *args, **kwargs)

    return wrapper


def requireAuthenticated(func: Callable):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.connect_info.sessionKey:
            raise InvaildSession("you must authenticate before this.")
        return func(self, *args, **kwargs)

    wrapper.__annotations__ = func.__annotations__
    return wrapper


def SinceVersion(*version: int):
    def wrapper(func):
        @functools.wraps(func)
        def inside_wrapper(self, *args, **kwargs):
            if (
                self.connect_info.current_version
                and self.connect_info.current_version < version
            ):
                raise NotSupportedVersion(
                    "the current version does not support this feature: {0}".format(
                        self.connect_info.current_version
                    )
                )
            return func(self, *args, **kwargs)

        return inside_wrapper

    return wrapper


def DeprecatedSince(*version: int, action: str = "warn"):
    if action not in ["warn", "error", "ignore"]:
        raise TypeError("action must be in" + str(["warn", "error", "ignore"]))

    def wrapper(func):
        @functools.wraps(func)
        def inside_wrapper(self, *args, **kwargs):
            if (
                self.connect_info.current_version
                and self.connect_info.current_version > version
            ):
                if action == "error":
                    raise NotSupportedVersion(
                        "the current version deprecated this feature: {0}".format(
                            self.connect_info.current_version
                        )
                    )
                elif action == "warn":
                    import warnings

                    warnings.warn(
                        "'{0}' has been deprecated since {1}, use other methods to realize your business as soon as possible!".format(
                            func.__qualname__, version
                        )
                    )
            return func(*args, **kwargs)

        return inside_wrapper

    return wrapper


code_exceptions_mapping = {
    1: InvaildAuthkey,
    2: AccountNotFound,
    3: InvaildSession,
    4: UnauthorizedSession,
    5: UnknownTarget,
    6: FileNotFoundError,
    10: PermissionError,
    20: AccountMuted,
    30: TooLongMessage,
    400: InvaildArgument,
}


def raise_for_return_code(code: Union[dict, int]):
    if isinstance(code, dict):
        code = code.get("code")
        exception_code = code_exceptions_mapping.get(code)
        if exception_code:
            raise exception_code
    elif isinstance(code, int):
        exception_code = code_exceptions_mapping.get(code)
        if exception_code:
            raise exception_code


def print_traceback_javay():
    stacks = inspect.stack()[1:]
    for i in stacks:
        print(f"    at [{i.filename}:{i.lineno}]")
    print("\n")


class AppMiddlewareAsDispatcher(BaseDispatcher):
    always = True
    context: ContextManager

    def __init__(self, app) -> None:
        self.app = app

    def beforeExecution(self, interface: "IDispatcherInterface"):
        self.context = enter_context(self.app, interface.event)
        self.context.__enter__()

    def afterExecution(self, interface: "IDispatcherInterface", exception, tb):
        self.context.__exit__(exception.__class__ if exception else None, exception, tb)

    def catch(self, interface: "DispatcherInterface"):
        pass


def context_enter_auto(context):
    def wrapper1(func):
        @functools.wraps(func)
        def wrapper2(*args, **kwargs):
            with context:
                return func(*args, **kwargs)

        return wrapper2

    return wrapper1


def call_atonce(*args, **kwargs):
    def wrapper(callable_target):
        return callable_target(*args, **kwargs)

    return wrapper


class InsertGenerator:
    base: Iterable[Any]
    insert_items: List[Any]

    def __init__(self, base_iterable: Iterable, pre_items: List[Any] = None) -> None:
        self.base = base_iterable
        self.insert_items = pre_items or []

    def __iter__(self):
        for i in self.base:
            if self.insert_items:
                yield from self.insert_items.pop()
            yield i
        else:
            if self.insert_items:
                for i in self.insert_items[::-1]:
                    yield from i


class AutoUnpackTuple:
    def __init__(self, base_iterable: Iterable, pre_items: List[Any] = None) -> None:
        self.base = base_iterable

    def __iter__(self):
        for i in self.base:
            if isinstance(i, tuple):
                yield from i
                continue
            yield i


def yes_or_no(value: bool) -> str:
    return "yes" if value else "no"



from enum import Enum
from pydantic import BaseModel


class UploadMethods(Enum):
    """用于向 `uploadImage` 或 `uploadVoice` 方法描述图片的上传类型"""

    Friend = "friend"
    Group = "group"
    Temp = "temp"


class MiraiConfig(BaseModel):
    """`/config` 接口的序列化实体类"""

    cacheSize: int
    enableWebsocket: bool



import copy
from typing import List, Sequence, Union
from .signature import FullMatch, PatternReceiver


class Arguments:
    content: List[PatternReceiver]

    @property
    def isGreed(self) -> bool:
        return any([i.isGreed for i in self.content])

    def __init__(self, content: List[PatternReceiver]) -> None:
        self.content = content

    def __repr__(self) -> str:
        return "<Argument content={0}>".format(str(self.content))


def merge_signature_chain_fullmatch(
    chain: Sequence[Union[FullMatch, PatternReceiver]]
) -> Sequence[Union[Arguments, FullMatch]]:
    result = []

    temp_l1 = []

    for i in chain:
        if isinstance(i, FullMatch):
            temp_l1.append(i.pattern)
        else:
            if temp_l1:
                result.append(FullMatch("".join(temp_l1)))
                temp_l1.clear()
            result.append(i)
    else:
        if temp_l1:
            result.append(FullMatch("".join(temp_l1)))
            temp_l1.clear()

    return type(chain)(result)


def merge_signature_chain(
    chain: Sequence[Union[FullMatch, PatternReceiver]]
) -> Sequence[Union[Arguments, FullMatch]]:
    chain = merge_signature_chain_fullmatch(chain)
    result = []
    temp_l1 = []

    for i in chain:
        if isinstance(i, PatternReceiver):
            temp_l1.append(i)
        else:
            if temp_l1:
                result.append(Arguments(temp_l1))
                temp_l1 = []
            result.append(i)
    else:
        if temp_l1:
            result.append(Arguments(temp_l1))
            temp_l1 = []

    return type(chain)(result)



from graia.broadcast import BaseDispatcher
from graia.broadcast.entities.event import BaseEvent
from . import ApplicationDispatcher, EmptyDispatcher


class SessionRefreshed(BaseEvent):
    "网络异常: 检测到无效的 Session 并自动刷新, 事件发布时已经刷新成为有效的 Session."

    Dispatcher = EmptyDispatcher


class SessionRefreshFailed(BaseEvent):
    "网络异常: 检测到无效的 Session 并尝试自动刷新失败."

    Dispatcher = EmptyDispatcher


class RemoteException(BaseEvent):
    "网络异常: 无头客户端处发生错误, 你应该检查其输出的错误日志."

    Dispatcher = EmptyDispatcher


class InvaildRequest(BaseEvent):
    "网络异常: 意料之外地, 发出了不被无头客户端接收的 HTTP 请求, 你应该通过相应渠道向我们汇报此问题"

    Dispatcher = EmptyDispatcher



import functools
from typing import Optional
from aiohttp import TraceConfig
from aiohttp.client import ClientSession
from aiohttp.tracing import TraceRequestEndParams, TraceRequestStartParams
from ..logger import AbstractLogger
import random
import string


class HttpRequestTracing:
    logger: AbstractLogger

    def __init__(self, logger: AbstractLogger) -> None:
        self.logger = logger

    async def on_request_start(
        self, session: ClientSession, trace_config_ctx, params: TraceRequestStartParams
    ):
        req_id = "".join(random.choices(string.ascii_letters, k=12))
        trace_config_ctx.req_id = req_id
        if params.method == "GET":
            self.logger.debug(f"Request[{req_id}:GET]: {params.url}")
        elif params.method == "POST":
            self.logger.debug(
                f"Request[{req_id}:POST]: {params.url} using [{trace_config_ctx.trace_request_ctx[1]['json'] if trace_config_ctx.trace_request_ctx else None}]"
            )

    async def on_request_end(
        self, session: ClientSession, trace_config_ctx, params: TraceRequestEndParams
    ):
        if params.method == "GET":
            self.logger.debug(f"Response[{trace_config_ctx.req_id}:GET]: {params.url}")
        elif params.method == "POST":
            self.logger.debug(
                f"Response[{trace_config_ctx.req_id}:POST]: {params.url} using [{trace_config_ctx.trace_request_ctx[1]['json'] if trace_config_ctx.trace_request_ctx else None}], responsed [{await params.response.text()}]"
            )

    def build_session(
        self, apply_for: Optional[ClientSession] = None, *args, **kwargs
    ) -> ClientSession:
        trace_config = TraceConfig()
        trace_config.on_request_start.append(self.on_request_start)
        trace_config.on_request_end.append(self.on_request_end)
        session = apply_for or ClientSession(
            trace_configs=[trace_config], *args, **kwargs
        )
        if apply_for is not None:
            session.trace_configs.append(trace_config)
            trace_config.freeze()

        def hook_method(method):
            @functools.wraps(method.__func__)
            def hooked_wrapper(*args, **kwargs):
                return method(*args, **{**kwargs, "trace_request_ctx": (args, kwargs)})

            return hooked_wrapper

        session.get = hook_method(session.get)
        session.post = hook_method(session.post)
        return session
